<?php

namespace Drupal\commerce_avatax\Plugin\Commerce\TaxType;

use Drupal\commerce_avatax\AvataxLib;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\Adjustment;
use Drupal\commerce_price\Price;
use Drupal\commerce_tax\Plugin\Commerce\TaxType\RemoteTaxTypeBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Provides the Avatax remote tax type.
 *
 * @CommerceTaxType(
 *   id = "avatax",
 *   label = "Avatax",
 * )
 */
class Avatax extends RemoteTaxTypeBase {

  /**
   * The Avatax library.
   *
   * @var \Drupal\commerce_avatax\AvataxLib
   */
  protected $avataxLib;

  /**
   * Constructs a new Avatax object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\commerce_avatax\AvataxLib $avatax_lib
   *   The Avatax library.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, EventDispatcherInterface $event_dispatcher, AvataxLib $avatax_lib) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $event_dispatcher);
    $this->avataxLib = $avatax_lib;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('event_dispatcher'),
      $container->get('commerce_avatax.avatax_lib')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'display_inclusive' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function apply(OrderInterface $order) {
    $response_body = $this->avataxLib->transactionsCreate($order);

    // Do not go further unless there have been lines added.
    if (empty($response_body['lines'])) {
      return;
    }
    $currency_code = $order->getTotalPrice() ? $order->getTotalPrice()->getCurrencyCode() : $order->getStore()->getDefaultCurrencyCode();
    $adjustments = [];
    $applied_adjustments = [];
    foreach ($response_body['lines'] as $tax_adjustment) {
      $adjustments[$tax_adjustment['lineNumber']] = $tax_adjustment['tax'];
    }

    // Add tax adjustments to order items.
    foreach ($order->getItems() as $item) {
      if (isset($adjustments[$item->uuid()])) {
        $item->addAdjustment(new Adjustment([
          'type' => 'tax',
          'label' => $this->t('Sales tax'),
          'amount' => new Price((string) $adjustments[$item->uuid()], $currency_code),
          'source_id' => $this->pluginId . '|' . $this->entityId,
        ]));
        $applied_adjustments[$item->uuid()] = $item->uuid();
      }
    }

    // If we still have Tax adjustments to apply, apply one to the order.
    $remaining_adjustments = array_diff_key($adjustments, $applied_adjustments);
    if ($remaining_adjustments) {
      $order->addAdjustment(new Adjustment([
        'type' => 'tax',
        'label' => $this->t('Sales tax'),
        'amount' => new Price((string) array_sum($remaining_adjustments), $currency_code),
        'source_id' => $this->pluginId . '|' . $this->entityId,
      ]));
    }
  }

}
